#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2022 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
import os
from contextlib import contextmanager
from enum import Enum
from functools import wraps
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Dict
from typing import NoReturn
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import TypeVar
from typing import Union

from rfc3986 import normalizers
from rfc3986 import ParseResult


LOG = logging.getLogger(__name__)


# ========================================================================= #
# PATCH                                                                     #
# ========================================================================= #


_ORIG_REMOVE_DOT_SEGMENTS = normalizers.remove_dot_segments


@contextmanager
def _rfc3986_patch_context__remove_dot_segments(disabled=False):
    # set new function, this no longer matches: http://tools.ietf.org/html/rfc3986#section-5.2.4
    # -- make sure that '..' and '.' at the start of a path are not removed!
    # -- '' might become '.' which should actually not be allowed!
    if not disabled:
        normalizers.remove_dot_segments = os.path.normpath
    # move into context
    try:
        yield
    # restore original function
    finally:
        normalizers.remove_dot_segments = _ORIG_REMOVE_DOT_SEGMENTS


# ========================================================================= #
# URI validation                                                            #
# ========================================================================= #


class UriMalformedException(Exception):
    def __init__(self, parsed: ParseResult, msg: str):
        super().__init__(f'[Malformed URI]: {msg} -- From URI: {parsed.unsplit()}')


class EnumValMode(Enum):
    OPTIONAL = 'OPTIONAL'
    REQUIRED = 'REQUIRED'
    FORBIDDEN = 'FORBIDDEN'


class UriFieldValidator(object):
    def __init__(
        self,
        mode: EnumValMode = EnumValMode.OPTIONAL,
        validator: Callable[[ParseResult, str, str, Any], ParseResult] = None,
        one_of: Optional[Sequence[Any]] = None,
    ):
        self._mode = mode
        self._validator = validator
        self._one_of = one_of

    def __call__(self, parsed: ParseResult, uri_kind: str, field_name: str, field_value: Any) -> NoReturn:
        # validate based on the mode
        if self._mode == EnumValMode.REQUIRED:
            if not field_value:
                raise UriMalformedException(parsed, f'field {repr(field_name)} is required, but got value: {repr(field_value)}')
        elif self._mode == EnumValMode.FORBIDDEN:
            if field_value:
                raise UriMalformedException(parsed, f'field {repr(field_name)} is forbidden, but got value: {repr(field_value)}')
        elif self._mode != EnumValMode.OPTIONAL:
            raise NotImplementedError('This should never happen!')
        # validate based on required values
        if self._one_of is not None:
            if field_value not in self._one_of:
                raise UriMalformedException(parsed, f'field {repr(field_name)} has value: {repr(field_value)}, but must be one of: {list(self._one_of)}')
        # validate based on the validator function
        if self._validator is not None:
            self._validator(parsed, uri_kind, field_name, field_value)


class UriValidator(object):
    # override these in subclasses
    validate_scheme:   UriFieldValidator = UriFieldValidator(mode=EnumValMode.OPTIONAL)
    validate_userinfo: UriFieldValidator = UriFieldValidator(mode=EnumValMode.OPTIONAL)
    validate_host:     UriFieldValidator = UriFieldValidator(mode=EnumValMode.OPTIONAL)
    validate_port:     UriFieldValidator = UriFieldValidator(mode=EnumValMode.OPTIONAL)
    validate_path:     UriFieldValidator = UriFieldValidator(mode=EnumValMode.OPTIONAL)
    validate_query:    UriFieldValidator = UriFieldValidator(mode=EnumValMode.OPTIONAL)
    validate_fragment: UriFieldValidator = UriFieldValidator(mode=EnumValMode.OPTIONAL)

    def __call__(self, uri: Union[str, Path]) -> ParseResult:
        return self.validate(uri)

    def validate(self, uri: Union[str, Path]) -> ParseResult:
        parsed = uri_parse(uri)
        # validate everything
        self.validate_scheme  (parsed=parsed, uri_kind=self.uri_kind, field_name='scheme',   field_value=parsed.scheme)
        self.validate_userinfo(parsed=parsed, uri_kind=self.uri_kind, field_name='userinfo', field_value=parsed.userinfo)
        self.validate_host    (parsed=parsed, uri_kind=self.uri_kind, field_name='host',     field_value=parsed.host)
        self.validate_port    (parsed=parsed, uri_kind=self.uri_kind, field_name='port',     field_value=parsed.port)
        self.validate_path    (parsed=parsed, uri_kind=self.uri_kind, field_name='path',     field_value=parsed.path)
        self.validate_query   (parsed=parsed, uri_kind=self.uri_kind, field_name='query',    field_value=parsed.query)
        self.validate_fragment(parsed=parsed, uri_kind=self.uri_kind, field_name='fragment', field_value=parsed.fragment)
        # final result
        return parsed

    # override these

    @property
    def uri_kind(self) -> str:
        raise NotImplementedError

    @property
    def uri_type(self) -> 'EnumUriType':
        raise NotImplementedError

    @property
    def allowed_schemes(self) -> Tuple[Optional[str], ...]:
        raise NotImplementedError

    @classmethod
    def extract(cls, validated: ParseResult) -> str:
        raise NotImplementedError


# ========================================================================= #
# URI Types                                                                 #
# ========================================================================= #


# including `str` as a parent class means we can write: `assert UriType.URL == 'URL'`
# without `str` as a parent class, this does not evaluate to `True`.
# -- HOWEVER: we do not want this feature! It will just cause confusion
#             and can't be type checked!
class EnumUriType(Enum):
    """
    Types of URIs supported by the `parse_uri_and_type` function.
    """
    FILE = 'FILE'
    URL = 'URL'
    S3 = 'S3'
    SSH = 'SSH'


class UriValidatorUrl(UriValidator):
    uri_kind = 'url'
    uri_type = EnumUriType.URL
    allowed_schemes = ('http', 'https')

    # override these in subclasses
    validate_scheme:   UriFieldValidator = UriFieldValidator(mode=EnumValMode.REQUIRED, one_of=allowed_schemes)
    validate_userinfo: UriFieldValidator = UriFieldValidator(mode=EnumValMode.FORBIDDEN)
    validate_host:     UriFieldValidator = UriFieldValidator(mode=EnumValMode.REQUIRED)
    validate_port:     UriFieldValidator = UriFieldValidator(mode=EnumValMode.OPTIONAL)
    validate_path:     UriFieldValidator = UriFieldValidator(mode=EnumValMode.OPTIONAL)
    validate_query:    UriFieldValidator = UriFieldValidator(mode=EnumValMode.OPTIONAL)
    validate_fragment: UriFieldValidator = UriFieldValidator(mode=EnumValMode.OPTIONAL)

    @classmethod
    def extract(cls, validated: ParseResult) -> str:
        return validated.geturl()


class UriValidatorFile(UriValidator):
    uri_kind = 'file'
    uri_type = EnumUriType.FILE
    allowed_schemes = ('file', None)

    # override these in subclasses
    validate_scheme:   UriFieldValidator = UriFieldValidator(mode=EnumValMode.OPTIONAL, one_of=allowed_schemes)
    validate_userinfo: UriFieldValidator = UriFieldValidator(mode=EnumValMode.FORBIDDEN)
    validate_host:     UriFieldValidator = UriFieldValidator(mode=EnumValMode.FORBIDDEN)
    validate_port:     UriFieldValidator = UriFieldValidator(mode=EnumValMode.FORBIDDEN)
    validate_path:     UriFieldValidator = UriFieldValidator(mode=EnumValMode.REQUIRED)
    validate_query:    UriFieldValidator = UriFieldValidator(mode=EnumValMode.FORBIDDEN)
    validate_fragment: UriFieldValidator = UriFieldValidator(mode=EnumValMode.FORBIDDEN)

    @classmethod
    def extract(cls, validated: ParseResult) -> str:
        return validated.path


# ========================================================================= #
# URI vars                                                                  #
# ========================================================================= #


_SCHEME_VALIDATORS: Dict[Optional[str], UriValidator] = {
    scheme: validate_cls()
    for validate_cls in [UriValidatorUrl, UriValidatorFile]
    for scheme in validate_cls.allowed_schemes
}


# ========================================================================= #
# URI validation                                                            #
# ========================================================================= #


def uri_parse(uri: Union[str, Path, ParseResult], rfc3986_norm: bool = False) -> ParseResult:
    parsed = uri
    # convert to parse result
    # -- assume already normalized if already a ParseResult
    if not isinstance(parsed, ParseResult):
        with _rfc3986_patch_context__remove_dot_segments(disabled=rfc3986_norm):
            parsed: ParseResult = ParseResult.from_string(str(parsed), lazy_normalize=False)
    # done!
    return parsed


def uri_validate(uri: Union[str, Path], return_validator: bool = False) -> Union[ParseResult, Tuple[ParseResult, UriValidator]]:
    parsed = uri_parse(uri)
    # get the validator
    validator = _SCHEME_VALIDATORS.get(parsed.scheme, None)
    if validator is None:
        raise KeyError(f'invalid uri scheme: {repr(parsed.scheme)}, must be one of: {list(_SCHEME_VALIDATORS.keys())}, for uri: {repr(parsed.geturl())}')
    # validate the uri
    validated = validator.validate(parsed)
    # get results
    if return_validator:
        return validated, validator
    return validated


def uri_extract(
    uri: Union[str, Path],
    return_validated: bool = False,
    return_validator: bool = False,
) -> Union[str, Tuple[str, ParseResult], Tuple[str, UriValidator], Tuple[str, ParseResult, UriValidator]]:
    validated, validator = uri_validate(uri, return_validator=True)
    # validate the uri
    uri_norm = validator.extract(validated)
    # return the single result
    if not (return_validator or return_validated):
        return uri_norm
    # return all the results
    results = [uri_norm]
    if return_validated: results.append(validated)
    if return_validator: results.append(validator)
    return tuple(results)


# ========================================================================= #
# URI Class - Errors                                                        #
# ========================================================================= #


T = TypeVar('T')


class UriIsIncorrectTypeError(Exception):
    """
    This error is thrown if the uri is an incorrect type
    """


def only_if(prop: property) -> Callable[[T], T]:
    def decorator(func: T) -> T:
        @wraps(func)
        def wrapper(self: 'Uri', *args, **kwargs):
            if getattr(self, prop.fget.__name__):
                raise UriIsIncorrectTypeError(f'Check if: `{prop.fget.__name__}` is `True` before calling `{func.__name__}`, got uri: {repr(self.uri)}')
            return func(self, *args, **kwargs)
        return wrapper
    return decorator


# ========================================================================= #
# URI Class                                                                 #
# ========================================================================= #


class Uri(object):

    def __init__(self, uri: Union[str, Path, ParseResult, 'Uri']):
        # unwrap uri object
        if isinstance(uri, Uri):
            uri = uri._input_uri
            assert not isinstance(uri, Uri)
        # save the input
        self._input_uri: Union[str, Path, ParseResult] = uri
        # get validated uri
        validated, validator = uri_validate(uri, return_validator=True)
        self._validated: ParseResult = validated
        self._validator: UriValidator = validator

    # ~=~=~ URI ~=~=~ #

    @property
    def uri_extract(self) -> str:
        return self._validator.extract(self._validated)

    @property
    def uri_type(self) -> EnumUriType:
        return self._validator.uri_type

    @property
    def uri_parsed(self) -> ParseResult:
        return self._validated

    @property
    def uri_basename(self) -> str:
        return os.path.basename(self._validated.path)

    @property
    def uri(self) -> str:
        return self._validated.geturl()

    def __repr__(self):
        return f'{self.__class__.__name__}(uri={repr(self._input_uri)})'

    def __str__(self):
        return self.uri

    # ~=~=~ FILE ~=~=~ #

    @property
    def is_file(self) -> bool:
        return self.uri_type == EnumUriType.FILE

    @property
    @only_if(is_file)
    def file(self) -> str:
        return self.uri_extract

    @property
    @only_if(is_file)
    def file_abs(self) -> str:
        return os.path.abspath(self.uri_extract)

    @property
    @only_if(is_file)
    def file_is_abs(self) -> bool:
        return os.path.isabs(self.uri_extract)

    # ~=~=~ URL ~=~=~ #

    @property
    def is_url(self) -> bool:
        return self.uri_type == EnumUriType.URL

    @property
    @only_if(is_url)
    def url(self) -> str:
        return self.uri_extract

    @property
    @only_if(is_url)
    def url_is_http(self) -> bool:
        return self.uri_parsed.scheme == 'http'

    @property
    @only_if(is_url)
    def url_is_https(self) -> bool:
        return self.uri_parsed.scheme == 'https'

    # ~=~=~ S3 ~=~=~ #

    @property
    def is_s3(self) -> bool:
        return self.uri_type == EnumUriType.S3

    # TODO ...

    # ~=~=~ SSH ~=~=~ #

    @property
    def is_ssh(self) -> bool:
        return self.uri_type == EnumUriType.SSH

    # TODO ...

    # ~=~=~ COMMON ~=~=~ #

    # def download(self, dst: Union[str, Path, 'Uri']):
    #     raise NotImplementedError
    #
    # def copy(self, dst: Union[str, Path, 'Uri']):
    #     raise NotImplementedError
    #
    # def retrieve(self, dst: Union[str, Path, 'Uri']):
    #     raise NotImplementedError


# ========================================================================= #
# export                                                                    #
# ========================================================================= #


__all__ = (
    # errors
    'UriMalformedException',
    'UriIsIncorrectTypeError',
    # enums
    'EnumValMode',
    'EnumUriType',
    # validation
    'UriFieldValidator',
    'UriValidator',
    'UriValidatorUrl',
    'UriValidatorFile',
    # functional
    'uri_parse',
    'uri_validate',
    'uri_extract',
    # oop
    'Uri',
)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #




#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
import os
import shutil
from pathlib import Path
from typing import BinaryIO
from typing import TextIO
from typing import Union

from doorway._modify_path import path_basename_modify


LOG = logging.getLogger(__name__)


# ========================================================================= #
# Atomic file saving                                                        #
# ========================================================================= #


_MODE_REPLACE = 'w'
_MODE_MISSING = 'x'
_MODE_TRY_COPY = 'a'
_MODE_EXISTING = 'r+'


class AtomicPath(object):
    """
    Within the context, data must be written to a temporary file.
    Once data has been successfully written, the temporary file
    is moved to the location of the destination file.

    The temporary path is set to the same directory as the
    destination path.

    ```
    with AtomicPath('file.txt') as tmp_path:
        with open(tmp_file, 'w') as f:
            f.write('hello world!\n')
    ```
    """

    # +----------------+-----+-----+-----+-----+-----+-----+-----+-----+
    # | DESCRIPTION    |  r  |  r+ |  w  |  w+ |  a  |  a+ |  x  |  x+ |
    # |----------------+-----+-----+-----+-----+-----+-----+-----+-----|
    # | read           |  *  |  *  |     |  *  |     |  *  |     |  *  |
    # | write          |     |  *  |  *  |  *  |  *  |  *  |  *  |  *  |
    # |----------------+-----+-----+-----+-----+-----+-----+-----+-----|
    # | create         |     |     |  *  |  *  |  *  |  *  |  *  |  *  |
    # | truncate       |     |     |  *  |  *  |     |     |  O  |  O  |
    # |----------------+-----+-----+-----+-----+-----+-----+-----+-----|
    # | must exist     |  *  |  *  |     |     |     |     |     |     |
    # | overwrite      |     |     |  *  |  *  |     |     |     |     |
    # | append         |     |     |     |     |  *  |  *  |     |     |
    # | must not exist |     |     |     |     |     |     |  *  |  *  |
    # |----------------+-----+-----+-----+-----+-----+-----+-----+-----|
    # | seek to start  |  *  |  *  |  *  |  *  |     |     |  *  |  *  |
    # | seek to end    |     |     |     |     |  *  |  *  |     |     |
    # |----------------+-----+-----+-----+-----+-----+-----+-----+-----|
    # | O: special case, must not exist, so truncated by default       |
    # +----------------------------------------------------------------+

    # SUPPORTED MODES:
    # 'r' : overwrite, assert temp file created in context and then move to destination
    #     | like open(..., 'w'): write which truncates the file
    # 'x' : exclusive, assert temp file created in context but fail if dest. exists
    #     | like open(..., 'x'): exclusive create which fails if the file already exists
    # 'a' : append, copy data to temp file if it exists, then move back to destination
    #     | like open(..., 'a'): append which appends to end of file if it exists
    # 'r+': required, copy data to temp file, then move back to destination
    #     | like open(..., 'r+'): read which requires the file to exist

    # UNSUPPORTED MODES:
    # 'r': open for reading
    # 'U': universal newlines mode

    def __init__(
        self,
        file: Union[str, Path],
        mode: str = _MODE_MISSING,
        makedirs: bool = False,
    ):
        # check files
        if (not file) or Path(file).name in ('', '.', '..'):
            raise ValueError(f'file must not be empty: {repr(file)}')

        # check the mode
        if mode not in (_MODE_TRY_COPY, _MODE_REPLACE, _MODE_MISSING, _MODE_EXISTING):
            raise ValueError(
                f'invalid mode: {repr(mode)}, '
                f'must be one of: {_MODE_TRY_COPY}/{_MODE_REPLACE}/{_MODE_MISSING}/{_MODE_EXISTING} (try_copy/replace/missing/existing)'
            )

        # get the actual files
        from uuid import uuid4
        self._dst_path = Path(file).absolute()
        self._tmp_path = path_basename_modify(
            file=self._dst_path,
            basename_prefix=f'.temp.{uuid4()}.',
        )

        # check that the files are different, but that
        # their parent directories are the same
        if self._dst_path == self._tmp_path:
            raise ValueError(f'temporary and destination files are the same: {self._tmp_path} == {self._dst_path}')
        if self._dst_path.parent != self._tmp_path.parent:
            raise ValueError(f'temporary and destination directories are not same: {self._tmp_path.parent} != {self._dst_path.parent}')

        # other settings
        self._makedirs = makedirs
        self._mode = mode

    def __enter__(self) -> Path:
        # 1. check that the temporary file does not already exist
        #    this should be impossible
        if self._tmp_path.exists():
            raise RuntimeError(f'the temporary file already exists: {self._tmp_path}, this is a bug!')

        # 2. handle the different modes for when the destination file exists
        # - make sure the destination does not exist
        if self._mode == _MODE_MISSING:
            if self._dst_path.exists():
                raise FileExistsError(f'the destination file should not exist: {self._dst_path}')
        # - make sure the destination can be replaced
        elif self._mode == _MODE_REPLACE:
            if self._dst_path.exists():
                if not self._dst_path.is_file():
                    raise IsADirectoryError(f'the destination file exists but is not a file: {self._dst_path}')
        # - make sure the destination can be replaced and try copy it
        elif self._mode in (_MODE_REPLACE, _MODE_TRY_COPY, _MODE_EXISTING):
            if self._dst_path.exists():
                if not self._dst_path.is_file():
                    raise IsADirectoryError(f'the destination file exists but is not a file: {self._dst_path}')
                shutil.copy(self._dst_path, self._tmp_path)
        # - make sure the destination exists, can be replaced and copy it
        elif self._mode == _MODE_EXISTING:
            if not self._dst_path.exists():
                raise FileExistsError(f'the destination file should exist: {self._dst_path}')
            elif not self._dst_path.is_file():
                raise FileExistsError(f'the destination file exists but is not a file: {self._dst_path}')
            shutil.copy(self._dst_path, self._tmp_path)
        # - make sure the mode is valid
        else:
            raise NotImplementedError(f'invalid mode: {self._mode}, this is a bug!')

        # 3. create the missing parent directory if specified
        if self._makedirs:
            self._tmp_path.parent.mkdir(parents=True, exist_ok=True)

        # return the path to the temp file
        return self._tmp_path

    def __exit__(self, error_type, error, traceback):
        # 0. cleanup if there was an error, and exit early
        if error_type is not None:
            if self._tmp_path.exists():
                if self._tmp_path.is_dir():
                    raise RuntimeError(f'An error occured in {self.__class__.__name__}, but could not clean up the temporary file because it is a directory: {self._tmp_path}')
                self._tmp_path.unlink(missing_ok=True)
                LOG.error(f'An error occurred in {self.__class__.__name__}, deleted temporary file: {self._tmp_path}')
            else:
                LOG.error(f'An error occurred in {self.__class__.__name__}')
            return

        # 1. check that the temporary file was created in this context
        if not self._tmp_path.exists():
            raise FileNotFoundError(f'the temporary file was not created: {self._tmp_path}')
        if not self._tmp_path.is_file():
            raise RuntimeError(f'the temporary file is not a file: {self._tmp_path}')

        # 2. handle the different modes, we perform some checks again just to be safe
        # - make sure the destination does not exist
        if self._mode == _MODE_MISSING:
            if self._dst_path.exists():
                raise FileExistsError(f'the destination file should not exist: {self._dst_path}')
        # - no checks needed
        elif self._mode in (_MODE_REPLACE, _MODE_TRY_COPY, _MODE_EXISTING):
            pass
        # - make sure the mode is valid
        else:
            raise NotImplementedError(f'invalid mode: {self._mode}, this is a bug!')

        # 3. move the temp file to the destination file. `os.rename` is usually
        # guaranteed to be atomic on linux and also overwrites the destination path
        LOG.info(f'moving temporary file to final location: {self._tmp_path} -> {self._dst_path}')
        os.rename(self._tmp_path, self._dst_path)


class AtomicOpen(object):

    # SUPPORTED MODES:
    # 'r': open for reading (default)
    # 'a': open for writing, appending to the end of the file if it exists
    # 'w': open for writing, truncating the file first
    # 'x': open for exclusive creation, failing if the file already exists

    # SUPPORTED MODIFIERS:
    # 't': text mode (default)
    # '+': open a disk file for updating (reading and writing)
    # 'b': binary mode

    # UNSUPPORTED MODES:
    # 'U': universal newlines mode (deprecated)

    def __init__(
        self,
        file: Union[str, Path],
        mode: str = 'x',
        makedirs: bool = False,
    ):
        # obtain the basic mode from the actual mode
        if 'r' in mode:
            basic_mode = _MODE_EXISTING if ('+' in mode) else None
        elif 'x' in mode:
            basic_mode = _MODE_MISSING
        elif 'w' in mode:
            basic_mode = _MODE_REPLACE
        elif 'a' in mode:
            basic_mode = _MODE_TRY_COPY
        else:
            raise ValueError(f'invalid mode: {repr(mode)}, must contain: r/x/w/a')

        # set the class vars
        self._open_mode = mode
        self._file_io = None
        self._orig_path = file

        # handle the different basic modes
        if basic_mode is None:
            self._atomic_path = None
        else:
            self._atomic_path = AtomicPath(
                file=file,
                mode=basic_mode,
                makedirs=makedirs,
            )

    def __enter__(self) -> Union[TextIO, BinaryIO]:
        # - we should be in read-only mode
        if self._atomic_path is None:
            tmp_path = self._orig_path
            LOG.debug(f'opening original file: {tmp_path} with mode: {self._open_mode}')
        # - prepare like usual
        else:
            tmp_path = self._atomic_path.__enter__()
            LOG.debug(f'opening temporary file: {tmp_path} with mode: {self._open_mode}')

        # open and return the file
        self._file_io = open(tmp_path, self._open_mode)
        return self._file_io

    def __exit__(self, error_type, error, traceback):
        # close the temp file
        try:
            self._file_io.close()
        finally:
            self._file_io = None
        # cleanup like usual if we are not in read-only mode
        if self._atomic_path is not None:
            self._atomic_path.__exit__(error_type, error, traceback)


# ========================================================================= #
# export                                                                    #
# ========================================================================= #


__all__ = (
    'AtomicPath',
    'AtomicOpen',
)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2022 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import hashlib
import os
import warnings
from pathlib import Path
from typing import Dict
from typing import Iterable
from typing import NoReturn
from typing import Optional
from typing import Union
from doorway._utils import VarHandlerStr


# ========================================================================= #
# byte producers                                                            #
# ========================================================================= #


def _yield_file_bytes(file: str, chunk_size=16384):
    with open(file, 'rb') as f:
        while True:
            bytes = f.read(chunk_size)
            if not bytes:
                return
            yield bytes


def _yield_fast_hash_bytes(file: str, chunk_size=16384, num_chunks=3):
    assert num_chunks >= 2
    # return the size in bytes
    size = os.path.getsize(file)
    yield size.to_bytes(length=64//8, byteorder='big', signed=False)
    # return file bytes chunks
    if size < chunk_size * num_chunks:
        # we can't return chunks because the file is too small, return everything!
        yield from _yield_file_bytes(file, chunk_size=chunk_size)
    else:
        # includes evenly spaced start, middle and end chunks
        with open(file, 'rb') as f:
            for i in range(num_chunks):
                pos = (i * (size - chunk_size)) // (num_chunks - 1)
                f.seek(pos)
                yield f.read(chunk_size)


# ========================================================================= #
# hash algos                                                                #
# ========================================================================= #


Hash = str
Hashes = Union[str, Dict[str, str]]
HashMode = str
HashAlgo = str
HashPath = Union[str, Path]


# ========================================================================= #
# hash mode                                                                 #
# ========================================================================= #


_FILE_BYTE_PRODUCERS = {
    'full': _yield_file_bytes,
    'fast': _yield_fast_hash_bytes,
}


_VAR_HANDLER_HASH_MODE = VarHandlerStr(
    identifier='hash_mode',
    environ_key='DOORWAY_HASH_MODE',
    fallback_value='fast',
    allowed_values=tuple(_FILE_BYTE_PRODUCERS.keys()),
)


def hash_mode_set_default(hash_mode: Optional[HashMode]) -> NoReturn:
    return _VAR_HANDLER_HASH_MODE.set_default_value(value=hash_mode)


def hash_mode_get(hash_mode: Optional[HashMode] = None) -> HashMode:
    return _VAR_HANDLER_HASH_MODE.get_value(override=hash_mode)


_VAR_HANDLER_HASH_ALGO = VarHandlerStr(
    identifier='hash_algo',
    environ_key='DOORWAY_HASH_ALGO',
    fallback_value='md5',
    allowed_values=tuple(hashlib.algorithms_guaranteed),  # hashlib.algorithms_available?
)


def hash_algo_set_default(hash_algo: Optional[HashAlgo]) -> NoReturn:
    return _VAR_HANDLER_HASH_ALGO.set_default_value(value=hash_algo)


def hash_algo_get(hash_algo: Optional[HashAlgo] = None) -> HashAlgo:
    return _VAR_HANDLER_HASH_ALGO.get_value(override=hash_algo)


# ========================================================================= #
# file hashing                                                              #
# ========================================================================= #


def hash_bytes(bytes_str: bytes, hash_algo: Optional[HashAlgo] = None) -> str:
    # normalise the hash_algo
    hash_algo = hash_algo_get(hash_algo=hash_algo)
    # generate hash and convert to a string
    return hashlib.new(hash_algo, data=bytes_str).hexdigest()


def hash_bytes_iter(bytes_iter: Iterable[bytes], hash_algo: Optional[HashAlgo] = None) -> str:
    # normalise the hash_algo
    hash_algo = hash_algo_get(hash_algo=hash_algo)
    # generate hash and convert to a string
    hash = hashlib.new(hash_algo)
    for bytes in bytes_iter:
        hash.update(bytes)
    return hash.hexdigest()


def hash_str(str: str, hash_algo: Optional[HashAlgo] = None, encoding: str = 'utf-8') -> str:
    # encode string as bytes and then hash
    return hash_bytes(str.encode(encoding), hash_algo=hash_algo)


def hash_file(
    path: HashPath,
    hash_mode: Optional[HashMode] = None,
    hash_algo: Optional[HashAlgo] = None,
    hash_missing: bool = False,
) -> Hash:
    """
    :param path: the path to the file
    :param hash_mode: "full" uses all the bytes in the file to compute the hash, "fast" uses the start, middle, end bytes as well as the size of the file in the hash. Default is "fast".
    :param hash_algo: the kind of hash algorithm to use, see `hashlib` for details. Default is "md5"
    :param hash_missing: If enabled, then an error is not thrown if the file is missing, rather an empty hash is returned!
    :return: the hexdigest of the hash
    :raises FileNotFoundError
    """
    # normalise the hash_mode
    hash_mode = hash_mode_get(hash_mode=hash_mode)
    # check the file exists
    path = str(path)
    if os.path.exists(path):
        if not os.path.isfile(path):
            raise IsADirectoryError(f'the path exists but is not a file: {repr(path)}')
    else:
        if hash_missing:
            return ''
        raise FileNotFoundError(f'could not compute hash for missing file: {repr(path)}')
    # get file bytes iterator
    byte_producer = _FILE_BYTE_PRODUCERS[hash_mode]
    bytes_iter = byte_producer(path)
    # get file bytes iterator
    return hash_bytes_iter(bytes_iter, hash_algo=hash_algo)


# ========================================================================= #
# file hashing utils                                                        #
# ========================================================================= #


class HashError(Exception):
    """
    Raised if the hash of a file was invalid.
    """


def hash_norm(
    hash: Hashes,
    hash_mode: Optional[HashMode] = None,
    hash_algo: Optional[HashAlgo] = None,
) -> Hash:
    """
    file hashes depend on the mode.
    - Allow hashes to be dictionaries that map the hash_mode or hash_algo to the hash.
      This function returns the correct hash if it is a dictionary.

    priority of keys:
       1. `mode:algo`
       2. `mode`
       3. `algo`
    """
    if isinstance(hash, dict):
        hash_mode = hash_mode_get(hash_mode)
        hash_algo = hash_algo_get(hash_algo)
        multi_key = f'{hash_mode}:{hash_algo}'
        # 1. try get `mode:algo`
        if multi_key in hash:
            hash = hash[multi_key]
        elif hash_mode in hash:
            warnings.warn('obtaining the hash directly from the `hash_mode` is deprecated, please use the full key `hash_mode:hash_algo`')
            hash = hash[hash_mode]
        elif hash_algo in hash:
            warnings.warn('obtaining the hash directly from the `hash_algo` is deprecated, please use the full key `hash_mode:hash_algo`')
            hash = hash[hash_algo]
        else:
            raise KeyError(f'hash dictionary does not contain a valid key for either 1. {repr(multi_key)}, 2. {repr(hash_mode)}, or 3. {repr(hash_algo)}. Available hash keys are: {sorted(hash.keys())}')
    # check the result
    if not isinstance(hash, str):
        raise TypeError(f'normalized hash should be a str, got type: {type(hash)} for value: {repr(hash)}')
    # done!
    return hash


def hash_file_validate(
    path: HashPath,
    hash: Hashes,
    hash_mode: Optional[HashMode] = None,
    hash_algo: Optional[HashAlgo] = None,
    hash_missing: bool = False,
) -> NoReturn:
    """
    :raises FileNotFoundError, HashError
    """
    # normalize the hash
    hash: str = hash_norm(hash=hash, hash_mode=hash_mode)
    # compute the hash
    fhash = hash_file(path=path, hash_algo=hash_algo, hash_mode=hash_mode, hash_missing=hash_missing)
    # check the hash
    if fhash != hash:
        # functions above also call this, we need to do it again for the error message
        hash_mode = hash_mode_get(hash_mode)
        hash_algo = hash_algo_get(hash_algo)
        # raise the error!
        raise HashError(f'computed {hash_mode} {hash_algo} hash: {repr(fhash)} does not match expected hash: {repr(hash)} for file: {repr(path)}')


def hash_file_is_valid(
    path: HashPath,
    hash: Hashes,
    hash_mode: Optional[HashMode] = None,
    hash_algo: Optional[HashAlgo] = None,
    hash_missing: bool = False,
) -> bool:
    try:
        hash_file_validate(path=path, hash=hash, hash_algo=hash_algo, hash_mode=hash_mode, hash_missing=hash_missing)
    except HashError:
        return False
    return True


# ========================================================================= #
# export                                                                    #
# ========================================================================= #


__all__ = (
    # types
    'Hash',
    'Hashes',
    'HashMode',
    'HashAlgo',
    'HashPath',
    # errors
    'HashError',
    # hash mode
    'hash_mode_set_default',
    'hash_mode_get',
    'hash_algo_set_default',
    'hash_algo_get',
    # normalise hash
    'hash_norm',
    # compute hash
    'hash_bytes',
    'hash_bytes_iter',
    'hash_str',
    'hash_file',
    'hash_file_validate',
    'hash_file_is_valid',
)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import os
from logging import getLogger
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple


_LOGGER = getLogger(__name__)


# ============================================================================ #
# Proxy Scraping                                                               #
# ============================================================================ #


_PROXY_SOURCES = {}
_DEFAULT_SOURCE = None


def set_default_proxy_scraper(name: str):
    if name not in _PROXY_SOURCES:
        raise KeyError(f'Cannot set as default! Scrape function with name: {repr(name)} does not exist.')
    global _DEFAULT_SOURCE
    if _DEFAULT_SOURCE != name:
        _LOGGER.info(f'overridden default proxy scrape_fn: {repr(_DEFAULT_SOURCE)} -> {repr(name)}')
        _DEFAULT_SOURCE = name


def register_proxy_scraper(name: str, scrape_fn: Optional[Callable[[str], List[Dict[str, str]]]] = None, is_default: bool = False):
    if name in _PROXY_SOURCES:
        raise KeyError('scrape function with name: {repr(name)} already exists')
    # decorator
    def wrapper(scrape_fn: Callable[[str], List[Dict[str, str]]]):
        # just in case the decorator call was delayed
        assert name not in _PROXY_SOURCES
        _PROXY_SOURCES.setdefault(name, scrape_fn)
        _LOGGER.debug(f'registered proxy scrape_fn: {repr(name)}')
        # set the default
        if is_default:
            set_default_proxy_scraper(name=name)
    # decorator or function
    if scrape_fn is None:
        return wrapper
    else:
        wrapper(scrape_fn)


def scrape_proxies(source: Optional[str] = None, proxy_type: str = 'all', cache_dir: str = 'data/proxies/cachier', cached: bool = True) -> List[Dict[str, str]]:
    if source is None:
        if _DEFAULT_SOURCE is None:
            raise RuntimeError('no default proxy scrape function has been set.')
        source = _DEFAULT_SOURCE
        _LOGGER.info(f'using default proxy scrape function: {repr(source)}')
    # get source
    try:
        proxy_scrape_fn = _PROXY_SOURCES[source]
    except KeyError:
        raise KeyError(f'proxy scrape function with name: {repr(source)} does not exist. Valid scrape sources are: {sorted(_PROXY_SOURCES.keys())}')
    # wrap the function
    if cached:
        from cachier import cachier
        from datetime import timedelta
        proxy_scrape_fn = cachier(
            stale_after=timedelta(days=1),
            backend='pickle',
            cache_dir=cache_dir
        )(proxy_scrape_fn)
    # obtain the proxies
    _LOGGER.info(f'scraping proxies from source: {repr(source)}')
    proxy_list = proxy_scrape_fn(proxy_type=proxy_type)
    _LOGGER.info(f'scrapped: {len(proxy_list)} proxies from source: {repr(source)}')
    # done!
    return proxy_list


# ============================================================================ #
# Proxy Scrapers                                                               #
# ============================================================================ #


def _requests_get(url, fake_user_agent=True, params=None):
    import requests
    # fake a request from a browser
    return requests.get(
        url,
        headers={} if not fake_user_agent else {'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/90.0.4430.212 Safari/537.36'},
        params=params,
    )


@register_proxy_scraper('morph.io')
def _scrape_proxies_morph(proxy_type) -> List[Dict[str, str]]:

    assert 'MORPH_API_KEY' in os.environ, 'MORPH_API_KEY environment variable not set!'
    morph_api_key = os.environ['MORPH_API_KEY']
    morph_api_url = "https://api.morph.io/CookieMichal/us-proxy/data.json"

    query = f"select * from 'data' where (anonymity='elite proxy' or anonymity='anonymous')"

    if 'https' == proxy_type:
        query += " and https='yes'"
    elif 'http' == proxy_type:
        query += " and https='no'"
    elif 'all' == proxy_type:
        pass
    else:
        raise KeyError(f'invalid proxy_type: {proxy_type}')

    r = _requests_get(
        morph_api_url,
        params={
            'key': morph_api_key,
            'query': query
        }
    )

    proxies = []
    for row in r.json():
        proto = 'HTTPS' if row['https'] == 'yes' else 'HTTP'
        url = "{}://{}:{}".format(proto, row['ip'], row['port'])
        proxies.append({proto: url})

    return proxies


@register_proxy_scraper('free-proxy-list.net', is_default=True)
def _scrape_proxies_freeproxieslist(proxy_type) -> List[Dict[str, str]]:
    def can_add(https):
        if proxy_type == 'all':
            return True
        elif proxy_type == 'https':
            return https == 'yes'
        elif proxy_type == 'http':
            return https == 'no'
        else:
            raise KeyError(f'invalid proxy_type: {proxy_type}')

    try:
        from bs4 import BeautifulSoup
    except:
        raise ImportError('BeautifulSoup `bs4` is not installed, cannot scrape proxies!')

    page = _requests_get('https://free-proxy-list.net/', fake_user_agent=True)
    soup = BeautifulSoup(page.content, 'html.parser')
    rows = soup.find_all('tr', recursive=True)

    proxies = []
    for row in rows:
        try:
            ip, port, country, country_long, anonymity, google, https, last_checked = (elem.text for elem in row.find_all('td', recursive=True))
            # check this entry is an ip entry
            if len(ip.split('.')) != 4:
                raise ValueError('not an ip entry')
            # filter entries
            if not can_add(https):
                continue
            # make entry
            proto = 'HTTPS' if (https == 'yes') else 'HTTP'
            url = "{}://{}:{}".format(proto, ip, int(port))
            proxies.append({proto: url})
        except:
            pass

    return proxies


# ============================================================================ #
# Proxy Errors                                                                 #
# ============================================================================ #


class MalformedProxyError(Exception):
    """
    raised if a proxy does not follow the correct format.
    eg. `proxy = {'<protocol>': '<protocol>://<url>'}`
    """

class NoMoreProxiesError(Exception):
    """
    raise if the ProxyDownloader has run out of proxies!
    """

class ProxyDownloadFailedError(Exception):
    """
    raise if the ProxyDownloader download has failed
    """


# ============================================================================ #
# Proxy Download Helper                                                        #
# ============================================================================ #


def make_proxy_opener(proxy: Dict[str, str]):
    import urllib.request

    if len(proxy) != 1:
        raise MalformedProxyError(f'proxy dictionaries should only have one entry, the key is the protocol, and the value is the url... invalid: {proxy}')
    # build connection
    return urllib.request.build_opener(
        urllib.request.ProxyHandler(proxy),
        urllib.request.ProxyBasicAuthHandler()
    )


def download_with_proxy(url: str, file: str, proxy: Dict[str, str], timeout: Optional[float] = 8):
    import io
    data = make_proxy_opener(proxy=proxy).open(url, timeout=timeout).read()
    # download to temp file in case there is an error
    temp_file = file + '.dl'
    with io.FileIO(temp_file, "w") as f:
        f.write(data)
    # make this atomic
    os.rename(temp_file, file)


def _skip_or_prepare_file(file: str, exists_mode: str, make_dirs: bool):
    """
    returns True if the file should be skipped, False otherwise.
    - also prepare the directories or deletion of the file!
    """
    if os.path.exists(file):
        # the file exists
        # make sure it is actually a file, not a directory or link
        if not os.path.isfile(file):
            raise IOError(f'the specified file is not a file: {file}')
        # handle the different modes
        if exists_mode == 'error':
            raise FileExistsError(f'the file already exists: {file}')
        elif exists_mode == 'skip':
            return True
        elif exists_mode == 'overwrite':
            os.unlink(file)
            _LOGGER.warning('overwriting file: {url}')
        else:
            raise KeyError(f'invalid exists_mode={repr(exists_mode)}')
    else:
        # the file does not exist
        # check the parent path
        parent_dir = os.path.dirname(file)
        if not os.path.exists(parent_dir):
            # the parent path does not exist
            if make_dirs:
                os.makedirs(parent_dir, exist_ok=True)
                _LOGGER.debug(f'[MADE] directory: {parent_dir}')
            else:
                raise FileNotFoundError(f'Parent directory does not exist: {parent_dir} Otherwise set make_dirs=True')
        else:
            # the parent path exists
            if not os.path.isdir(parent_dir):
                raise NotADirectoryError(f'Parent directory is not a directory: {parent_dir}')
    return False


# ============================================================================ #
# Proxy Downloader                                                             #
# ============================================================================ #


class ProxyDownloader:

    def __init__(
        self,
        proxies: Optional[Sequence[Dict[str, str]]],
        req_min_remove_count=5,
        req_max_fail_ratio=0.5,
    ):
        from collections import defaultdict
        from random import Random

        # default proxy scraping
        if proxies is None:
            proxies = scrape_proxies()
        # convert
        self._proxies = list(proxies)  # TODO: add support for raw proxy strings?
        # proxy statistics
        self._req_counts = defaultdict(int)
        self._req_fails = defaultdict(int)
        self._req_max_fail_ratio = req_max_fail_ratio
        self._req_min_remove_count = req_min_remove_count
        # random instance
        self._rand = Random()  # TODO: add round robbin mode?

    def random_proxy(self) -> Dict[str, str]:
        if len(self._proxies) <= 0:
            raise NoMoreProxiesError('The proxy downloader has run out of valid proxies.')
        # return a random proxy!
        index = self._rand.randint(0, len(self._proxies) - 1)
        return self._proxies[index]

    def _update_proxy(self, proxy: Dict[str, str], success: bool):
        (purl,) = proxy.values()
        # update uses and failures
        self._req_counts[purl] += 1
        self._req_fails[purl] += int(bool(not success))
        # make remove if there was an error
        counts, fails = self._req_counts[purl], self._req_fails[purl]
        if (counts > self._req_min_remove_count) and (fails / counts > self._req_max_fail_ratio):
            try:
                self._proxies.remove(proxy)
                del self._req_counts[purl]
                del self._req_fails[purl]
            except (ValueError, KeyError):
                pass  # removed in another thread

    def download_threaded(self, url_file_tuples: Sequence[Tuple[str, str]], exists_mode: str = 'error', verbose: bool = False, make_dirs: bool = False, ignore_failures=False, threads=64, attempts: int = 128, timeout: int = 8):
        from multiprocessing.pool import ThreadPool
        from tqdm import tqdm

        # check inputs
        if len(url_file_tuples) < 0:
            return []

        def download(url_file):
            url, file = url_file
            try:
                self.download(url=url, file=file, exists_mode=exists_mode, verbose=verbose, make_dirs=make_dirs, attempts=attempts, timeout=timeout)
            except ProxyDownloadFailedError:
                if ignore_failures:
                    return url, file
                else:
                    raise
            return None

        def get_desc():
            if ignore_failures:
                return f'Downloading [p={len(self._proxies)},t={threads},f={len(failed)}]'
            else:
                return f'Downloading [p={len(self._proxies)},t={threads}]'

        # download all files, keeping track of failed items!
        failed = []
        with ThreadPool(processes=threads) as pool:
            with tqdm(desc=get_desc(), total=len(url_file_tuples)) as pbar:
                for pair in pool.imap_unordered(download, url_file_tuples):
                    if pair:
                        failed.append(pair)
                    pbar.desc = get_desc()
                    pbar.update()

        # return all tuples for failed attempts
        return failed

    def download(self, url, file, exists_mode='error', verbose=False, make_dirs=False, attempts: int = 128, timeout: int = 8):
        """
        Download a file using random proxies.
        """
        if _skip_or_prepare_file(file=file, exists_mode=exists_mode, make_dirs=make_dirs):
            if verbose:
                _LOGGER.debug(f"[SKIPPED]: {file} | {url}")
            return
        # attempt download
        for i in range(attempts):
            proxy = self.random_proxy()
            try:
                download_with_proxy(url, file, proxy=proxy, timeout=timeout)
                if verbose:
                    _LOGGER.info(f"[DOWNLOADED]: {file} | {url}")
                self._update_proxy(proxy, success=True)
                return
            except Exception as e:
                if verbose:
                    _LOGGER.debug(f"[FAILED ATTEMPT {i + 1}]: {file} | {url} -- {e}")
                self._update_proxy(proxy, success=False)
        # download failed
        raise ProxyDownloadFailedError(f"[FAILED] tries={attempts}: {file} | {url}")


# ============================================================================ #
# Entrypoint                                                                   #
# ============================================================================ #


if __name__ == '__main__':

    def _command_line_app():
        import argparse
        import logging

        # parse arguments
        parser = argparse.ArgumentParser()
        parser.add_argument('-d', '--cache-dir', type=str, default='data/cache/proxies')
        parser.add_argument('-t', '--proxy-type', type=str, default='all')
        parser.add_argument('-s', '--proxy-source', type=str, default=None)
        parser.add_argument('-f', '--force-download', action='store_true')
        args = parser.parse_args()

        # download the proxies
        logging.basicConfig(level=logging.DEBUG)
        ProxyDownloader(
            proxies=scrape_proxies(
                source=args.proxy_source,
                proxy_type=args.proxy_type,
                cache_dir=args.cache_dir,
                cached=not args.force_download,
            )
        )

    _command_line_app()


# ============================================================================ #
# END                                                                          #
# ============================================================================ #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2022 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import os
from typing import NoReturn
from typing import Optional
from typing import Sequence
from typing import Generic
from typing import TypeVar


# ========================================================================= #
# Variable Manager                                                          #
# ========================================================================= #


T = TypeVar('T')


class VarHandlerBase(Generic[T]):

    def __init__(
        self,
        identifier: str,
        environ_key: str,
        fallback_value: T,
    ):
        self._identifier = identifier
        self._environ_key = environ_key
        assert str.isidentifier(self._environ_key)
        assert str.isidentifier(self._identifier)
        # default
        self._value_default = None
        # values
        self._value_fallback = fallback_value
        self._validate_value(self._value_fallback, source='fallback_value')

    # OVERRIDEABLE

    def _validate_value(self, value: T, source: str) -> NoReturn:  # pragma: no cover
        raise NotImplementedError

    def _normalize_environ_value(self, value: str) -> T:  # pragma: no cover
        raise NotImplementedError

    # COMMON - PROPS

    @property
    def identifier(self) -> str:
        return self._identifier

    @property
    def environ_key(self) -> str:
        return self._environ_key

    @property
    def fallback_value(self) -> T:
        return self._value_fallback

    # COMMON - FUNCS

    def set_default_value(self, value: Optional[T] = None) -> NoReturn:
        # make sure the hash_algo is valid
        if value is not None:
            self._validate_value(value, source='set_default_value')
        # update the default mode
        self._value_default = value

    def del_default_value(self) -> NoReturn:
        self._value_default = None

    def get_value(self, override: Optional[T] = None) -> T:
        """
        priority:
          1. manual specification
          2. default mode (set_default_value)
          3. environment variable
          4. fallback mode ("fast")
        """
        if override is not None:
            source, value = ('manual', override)
        elif self._value_default is not None:
            source, value = ('default', self._value_default)
        elif self._environ_key in os.environ:
            source, value = ('environment', self._normalize_environ_value(os.environ[self._environ_key]))
        else:
            source, value = ('fallback', self._value_fallback)
        # make sure the hash mode is valid
        self._validate_value(value=value, source=source)
        # done
        return value


# ========================================================================= #
# String Variable Manager                                                   #
# ========================================================================= #


# TODO: add handlers for different types, eg. bool
class VarHandlerStr(VarHandlerBase[str]):

    def __init__(
        self,
        identifier: str,
        environ_key: str,
        fallback_value: str,
        allowed_values: Sequence[str],
    ):
        # values
        self._allowed_values = set(allowed_values)
        # checks
        if len(self.allowed_values) <= 0:
            raise ValueError(f'allowed_values must not be an empty sequence, got: {repr(self._allowed_values)}')
        if not all(isinstance(v, str) for v in self._allowed_values):
            raise ValueError(f'all entries in the allowed_values must be strings, got: {repr(self._allowed_values)}')
        if fallback_value not in self._allowed_values:
            raise ValueError(f'the fallback_value: {repr(fallback_value)} is not one of the allowed_values: {repr(self._allowed_values)}')
        # initialize
        super().__init__(identifier=identifier, environ_key=environ_key, fallback_value=fallback_value)

    # CUSTOM

    @property
    def allowed_values(self) -> list:
        return sorted(self._allowed_values)

    # OVERRIDDEN

    def _validate_value(self, value: str, source: Optional[str] = None) -> NoReturn:
        if not isinstance(value, str):
            raise TypeError(f'invalid {self.identifier}: {repr(value)}, obtained from source: {source}, must be of type {str}, got type: {type(value)}')
        if value not in self._allowed_values:
            raise KeyError(f'invalid {self.identifier}: {repr(value)}, obtained from source: {source}, must be one of the allowed_values: {self.allowed_values}')

    def _normalize_environ_value(self, value: str) -> str:
        return value


# ========================================================================= #
# Bool Variable Manager                                                     #
# ========================================================================= #


class VarHandlerBool(VarHandlerBase[bool]):

    def __init__(
        self,
        identifier: str,
        environ_key: str,
        fallback_value: bool,
        environ_keys_true: Sequence[str] = ('y', 'yes', 't', 'true', '1'),
        environ_keys_false: Sequence[str] = ('n', 'no', 'f', 'false', '0'),
        environ_to_lower_case: bool = True,
    ):
        # values
        self._environ_keys_true = set(environ_keys_true)
        self._environ_keys_false = set(environ_keys_false)
        self._environ_to_lower_case = environ_to_lower_case
        # checks
        assert self._environ_keys_true and all(isinstance(v, str) for v in self._environ_keys_true)
        assert self._environ_keys_false and all(isinstance(v, str) for v in self._environ_keys_false)
        assert isinstance(environ_to_lower_case, bool)
        # init
        super().__init__(identifier=identifier, environ_key=environ_key, fallback_value=fallback_value)

    def _validate_value(self, value: bool, source: str) -> NoReturn:
        if not isinstance(value, bool):
            raise TypeError(f'invalid {self.identifier}: {repr(value)}, obtained from source: {source}, must be of type {bool}, got type: {type(value)}')

    def _normalize_environ_value(self, value: str) -> bool:
        if self._environ_to_lower_case:
            value = value.lower()
        if value in self._environ_keys_true:
            return True
        elif value in self._environ_keys_false:
            return False
        else:
            raise TypeError(f'cannot normalize environment variable `{self.environ_key}={repr(value)}` into {self.identifier}, must be one of: {sorted(self._environ_keys_true | self._environ_keys_false)}')


# ========================================================================= #
# export                                                                    #
# ========================================================================= #


__all__ = (
    'VarHandlerBase',
    'VarHandlerStr',
    'VarHandlerBool',
)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
from functools import wraps
from typing import Callable
from typing import NoReturn
from typing import Optional
from typing import Union

from doorway._hash import HashPath
from doorway._hash import Hashes
from doorway._hash import HashAlgo
from doorway._hash import HashMode
from doorway._hash import hash_file
from doorway._hash import hash_norm
from doorway._hash import hash_file_validate


LOG = logging.getLogger(__name__)


# ========================================================================= #
# Functional Stalefile                                                      #
# ========================================================================= #


def stalefile_is_stale(
    path: HashPath,
    hash: Hashes,
    hash_mode: Optional[HashMode] = None,
    hash_algo: Optional[HashAlgo] = None,
):
    """
    Check if the given path is stale:
    - a. if the file does not exist or if it's hash does not match then return `True`
    - b. if everything is okay then return `False`
    """
    # compute the hash for a file
    fhash = hash_file(path=path, hash_mode=hash_mode, hash_algo=hash_algo, hash_missing=True)
    # check if the file is stale or not
    if not fhash:
        LOG.info(f'file is stale because it does not exist: {repr(path)}')
        return True
    # obtain the target hash
    hash = hash_norm(hash=hash, hash_mode=hash_mode, hash_algo=hash_algo)
    # check if the file is stale or not
    if fhash != hash:
        LOG.warning(f'file is stale because the computed {hash_mode}:{hash_algo} hash: {fhash} does not match the target hash: {hash} for file: {repr(path)}')
        return True
    # the file is fresh
    LOG.debug(f'file is fresh: {repr(path)}')
    return False


def stalefile_generate(
    make_file_fn: Callable[[HashPath], NoReturn],
    path: HashPath,
    hash: Hashes,
    hash_mode: Optional[HashMode] = None,
    hash_algo: Optional[HashAlgo] = None,
) -> HashPath:
    """
    # if the file is stale:
    # - 1. produce the file using the wrapped function
    # - 2. validate the produced file and throw errors if it is wrong!
    # otherwise, do nothing.
    """
    is_stale = stalefile_is_stale(path=path, hash=hash, hash_mode=hash_mode, hash_algo=hash_algo)
    if is_stale:
        LOG.debug(f'calling wrapped function: {make_file_fn} because the file is stale: {repr(path)}')
        make_file_fn(path)
        hash_file_validate(path, hash=hash, hash_mode=hash_mode, hash_algo=hash_algo, hash_missing=True)
    # if the file is fresh
    # - 1. don't actually do anything, skip calling the producer!
    else:
        LOG.debug(f'skipped wrapped function: {make_file_fn} because the file is fresh: {repr(path)}')
    # return the path that contains the valid file!
    return path


def stalefile_decorator(
    path: str,
    hash: Hashes,
    hash_mode: Optional[HashMode] = None,
    hash_algo: Optional[HashAlgo] = None,
    make_file_fn: Optional[Callable[[HashPath], NoReturn]] = None,
) -> Union[
        Callable[[Callable[[HashPath], NoReturn]], Callable[[], HashPath]],
        Callable[[], HashPath],
]:
    """
    The wrapped function should take in a path and produce a file at that location.
    - a. if the file already exists, this function is not called!
    - b. if the file does not exist, the function is called to generate the file, which is then validated!
    """
    def decorator(make_file_fn: Callable[[HashPath], NoReturn]) -> Callable[[], HashPath]:
        @wraps(make_file_fn)
        def make_file_if_stale() -> HashPath:
            return stalefile_generate(
                make_file_fn=make_file_fn,
                path=path,
                hash=hash,
                hash_mode=hash_mode,
                hash_algo=hash_algo,
            )
        return make_file_if_stale
    # wrap directly if function is specified
    if make_file_fn is not None:
        return decorator(make_file_fn)
    return decorator


# ========================================================================= #
# Stalefile Class                                                           #
# ========================================================================= #


class Stalefile(object):
    """
    Common stalefile helper class that stores path and
    hash information, allowing re-use.

    See:
      - `stalefile_is_stale(...)` which corresponds to `self.is_stale()`
      - `stalefile_generate(...)` which corresponds to `self.generate(make_file_fn)`
      - `stalefile_decorator(...)` which corresponds to `self.decorator(make_file_fn)`
    """

    def __init__(
        self,
        path: str,
        hash: Hashes,
        hash_mode: Optional[HashMode] = None,
        hash_algo: Optional[HashAlgo] = None,
    ):
        self._path = path
        self._hash = hash
        self._hash_mode = hash_mode
        self._hash_algo = hash_algo

    def generate(self, make_file_fn: Callable[[HashPath], NoReturn]) -> HashPath:
        return stalefile_generate(
            make_file_fn=make_file_fn,
            path=self._path,
            hash=self._hash,
            hash_mode=self._hash_mode,
            hash_algo=self._hash_algo,
        )

    def decorator(self, make_file_fn: Optional[Callable[[HashPath], NoReturn]] = None) -> Callable[[], HashPath]:
        # the wrapped function should take in a path and produce a file at that location.
        # a. if the file already exists, this function is not called!
        # b. if the file does not exist, the function is called to generate the file, which is then validated!
        return stalefile_decorator(
            path=self._path,
            hash=self._hash,
            hash_mode=self._hash_mode,
            hash_algo=self._hash_algo,
            make_file_fn=make_file_fn,
        )

    def is_stale(self):
        return stalefile_is_stale(
            path=self._path,
            hash=self._hash,
            hash_mode=self._hash_mode,
            hash_algo=self._hash_algo,
        )

    def __bool__(self):
        return self.is_stale()


# ========================================================================= #
# export                                                                    #
# ========================================================================= #


__all__ = (
    'stalefile_is_stale',
    'stalefile_generate',
    'stalefile_decorator',
    'Stalefile',
)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



import os
from pathlib import Path
from typing import Callable
from typing import Iterable
from typing import Optional
from typing import Union
from typing import TypeVar

from doorway._hash import Hash
from doorway._hash import HashAlgo
from doorway._hash import hash_str


# ========================================================================= #
# individual shards                                                         #
# ========================================================================= #


T = TypeVar('T')

_SHARD_KEYS = {
    'basename': os.path.basename,
    'abspath': os.path.abspath,
    'input': lambda x: x,
}

ShardKey = Optional[Union[str, Callable[[T], str]]]


def shard_hash(
    value: T,
    shard_key: ShardKey[T] = None,
    hash_algo: Optional[HashAlgo] = None,
) -> Hash:
    # get the hash data function
    if shard_key is None:
        pass
    elif callable(shard_key):
        value = shard_key(value)
    elif isinstance(shard_key, str):
        fn = _SHARD_KEYS.get(shard_key, None)  # cannot be None here
        if fn is None:
            raise KeyError(f'if shard_key is a str, it must be one of: {list(_SHARD_KEYS.keys())}, got: {repr(shard_key)}')
        value = fn(value)
    else:
        raise ValueError(f'shard_key must be a str, callable or None, got: {repr(shard_key)}')
    # get the string
    assert isinstance(value, (str, Path)), f'The value after shard_key is applied must be a str or Path, instead got type: {type(value)}, with value: {repr(value)}'
    # compute the hash
    return hash_str(str(value), hash_algo=hash_algo)


def shard_idx(
    value: T,
    num_shards: int,
    shard_key: ShardKey[T] = None,
    hash_algo: Optional[HashAlgo] = None,
) -> int:
    assert isinstance(num_shards, int) and (num_shards > 0), f'num_shards must be an integer that is > 0, got: {repr(num_shards)}'
    # compute the hash for the value
    hash = shard_hash(value, shard_key=shard_key, hash_algo=hash_algo)
    # convert hashes to integers, and assign to correct split
    return int(hash, 16) % num_shards


# ========================================================================= #
# multiples shards                                                          #
# ========================================================================= #


_SHARD_RETURNS = {
    'pairs':   lambda i, value: (i, value),
    'indices': lambda i, value: i,
    'values':  lambda i, value: value,
}


def sharded(
    values: Iterable[T],
    num_shards: int,
    shard_key: ShardKey[T] = None,
    hash_algo: Optional[HashAlgo] = None,
    returns: str = 'values'
) -> list:
    """
    Shard files based on their hashes instead of random seeds
    """
    # shard functions
    if returns not in _SHARD_RETURNS:
        raise KeyError(f'invalid shards returns: {repr(returns)}, must be one of: {sorted(_SHARD_RETURNS.keys())}')
    value_getter = _SHARD_RETURNS[returns]
    # create new array of shards
    shards = [[] for _ in range(num_shards)]
    # assign paths to shards
    for i, value in enumerate(values):
        idx = shard_idx(value, num_shards, shard_key=shard_key, hash_algo=hash_algo)
        shards[idx].append(value_getter(i, value))
    # results
    return shards


def sharded_and_grouped(
    values: Iterable[T],
    group_sizes: Iterable[int],
    shard_key: ShardKey[T] = None,
    hash_algo: Optional[HashAlgo] = None,
    returns: str = 'values'
) -> list:
    """
    Shard files based on their hashes instead of random seeds
    -- This is useful if you need to split a dataset, but you expect changes to be made to it,
       eg. files will always randomly be assigned to the same shared
    """
    group_sizes = list(group_sizes)
    # checks
    assert all(isinstance(size, int) and (size >= 0) for size in group_sizes), f'values of group_sizes must be integers that are >= 0, got: {repr(group_sizes)}'
    # get all the shards
    shards = sharded(
        values=values,
        num_shards=sum(group_sizes),
        shard_key=shard_key,
        hash_algo=hash_algo,
        returns=returns,
    )
    # group all the shards together
    splits, i = [], 0
    for size in group_sizes:
        splits.append([path for shard in shards[i:i+size] for path in shard])
        i += size
    # done!
    return splits


# ========================================================================= #
# export                                                                    #
# ========================================================================= #


__all__ = (
    'shard_hash',
    'shard_idx',
    'sharded',
    'sharded_and_grouped',
)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
from doorway._atomic import AtomicOpen


LOG = logging.getLogger(__name__)


# ========================================================================= #
# files/dirs exist                                                          #
# ========================================================================= #


def io_download(
    src_url: str,
    dst_path: str,
    overwrite_existing: bool = False,
    chunk_size: int = 16384,
):
    # make sure we have the correct imports
    try:
        import requests
        from tqdm import tqdm
    except ImportError as e:
        raise ImportError(f'`requests` and `tqdm` need to be installed for `{io_download.__name__}`') from e

    # write the file
    with AtomicOpen(dst_path, 'wb' if overwrite_existing else 'xb') as fp:
        response = requests.get(src_url, stream=True)

        # get the file size from the request for the progress bar
        total_length = response.headers.get('content-length')
        if total_length is not None:
            total_length = int(total_length)

        # download with progress bar
        LOG.info(f'Downloading: {src_url} to: {dst_path}')
        with tqdm(total=total_length, desc=f'Downloading', unit='B', unit_scale=True, unit_divisor=1024) as progress:
            for data in response.iter_content(chunk_size=chunk_size):
                fp.write(data)
                progress.update(chunk_size)


# ========================================================================= #
# export                                                                    #
# ========================================================================= #


__all__ = (
    'io_download',
)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from pathlib import Path
from typing import Optional
from typing import Tuple
from typing import Union


# ========================================================================= #
# Basename formatting                                                       #
# ========================================================================= #


def basename_split_ext(
    basename: str,
    name_contains_sep: bool = True,
) -> Tuple[str, str]:
    # split the name from the basename
    if name_contains_sep:
        components = basename.rsplit('.', 1)
    else:
        components = basename.split('.', 1)
    # handle the case where there is no extension
    if len(components) == 1:
        return components[0], ''
    elif len(components) == 2:
        return components[0], f'.{components[1]}'
    else:
        raise NotImplementedError('this is a bug!')


def basename_modify(
    basename: str,
    ext: Optional[str] = None,
    name_prefix: Optional[str] = None,
    name_suffix: Optional[str] = None,
    basename_prefix: Optional[str] = None,
    basename_suffix: Optional[str] = None,
    name_contains_sep: bool = True,
) -> str:
    # 1. surround the name & replace the extension
    if name_suffix or ext:
        name, dot_ext = basename_split_ext(basename, name_contains_sep=name_contains_sep)
        if ext:
            dot_ext = f'.{ext}'
        if name_suffix:
            name = f'{name}{name_suffix}'
        basename = f'{name}{dot_ext}'
    if name_prefix:
        basename = f'{name_prefix}{basename}'
    # 2. surround the basename
    if basename_suffix:
        basename = f'{basename}{basename_suffix}'
    if basename_prefix:
        basename = f'{basename_prefix}{basename}'
    # done
    return basename


# ========================================================================= #
# Path Formatting                                                           #
# ========================================================================= #


def path_basename_modify(
    file: Union[str, Path],
    ext: Optional[str] = None,
    name_prefix: Optional[str] = None,
    name_suffix: Optional[str] = None,
    basename_prefix: Optional[str] = None,
    basename_suffix: Optional[str] = None,
    name_contains_sep: bool = True,
) -> Union[str, Path]:
    # get path components
    path = Path(file)
    basename = path.name
    if not basename:
        raise ValueError(
            f'file basename cannot be empty, '
            f'got basename: {repr(basename)}, '
            f'from file: {repr(str(file))}'
        )
    # update the basename
    basename = basename_modify(
        basename=basename,
        ext=ext,
        name_prefix=name_prefix,
        name_suffix=name_suffix,
        basename_prefix=basename_prefix,
        basename_suffix=basename_suffix,
        name_contains_sep=name_contains_sep,
    )
    # recombine the path and basename
    path = path.parent.joinpath(basename)
    # return path with same format as input
    return str(path) if isinstance(file, str) else path


# ========================================================================= #
# export                                                                    #
# ========================================================================= #


__all__ = (
    'basename_split_ext',
    'basename_modify',
    'path_basename_modify',
)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #
